#!/bin/sh
# STRandR (SpoopyTim RandR) - A utility to manage multiple displays using xrandr
# Required Packages: toml
# Designed and initially written internally at HorizonApp. Open-sourced for the greater good \(^o^)/

import subprocess
import argparse
import logging
import re
import sys
from typing import Generator, List, Dict

formatter = logging.Formatter('[%(asctime)s] [%(levelname)s] %(message)s', datefmt='%Y-%m-%d %H:%M:%S')

root_logger = logging.getLogger()
root_handler = logging.StreamHandler()
root_handler.setFormatter(formatter)
root_logger.addHandler(root_handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)

try:
    import toml
except ImportError:
    logger.fatal("You must install the 'toml' package for this script to work")
    exit(1)

class STRandR:
    def __init__(self):
        pass

    def get_xrandr_config(self):
        # Run xrandr command to get current configuration
        result = subprocess.run(['xrandr'], capture_output=True, text=True)
        return result.stdout

    def parse_xrandr_config(self, xrandr_output: str):
        displays = {}
        resolutions = {}

        display_pattern = re.compile(r'(\S+) connected.*?(\d+)x(\d+)\+(\d+)\+(\d+)')

        current_display = None
        display_resolutions = []
        for line in xrandr_output.splitlines():
            display_match = display_pattern.match(line)
            if display_match:
                display_name, width, height, x_pos, y_pos = display_match.groups()
                current_display = display_name
                displays[current_display] = {
                    'primary': ('primary' in line or False),
                    'enabled': True,
                    'mode': f'{width}x{height}',
                    'pos': f'{x_pos}x{y_pos}',
                    'scaling-override': ''
                }
                display_resolutions.append(f"{width}x{height}")
        for resolution in display_resolutions:
            resolutions[resolution] = {'scaling': '1x1'}

        return {'Displays': displays, 'Resolutions': resolutions}

    def get_current_config(self):
        xrandr_output = self.get_xrandr_config()
        parsed_config = self.parse_xrandr_config(xrandr_output)
        return toml.dumps(parsed_config)

    def configure_displays(self, config: Dict[str, Dict[str, str]]):
        commands = []
        for display_name, display_info in config['Displays'].items():
            logger.info(f"Display '{display_name}'")
            primary = display_info.get('primary', False)
            enabled = display_info.get('enabled', True)
            mode = display_info.get('mode', '')
            pos = display_info.get('pos', '')
            refresh_rate = display_info.get('refresh-rate', '')
            rotation = display_info.get('rotation', 'normal')
            scaling_override = display_info.get('scaling-override', '')
            scaling = scaling_override or config.get('Resolutions', {}).get(display_info.get('mode'), {}).get('scaling')
            
            logger.info(f"  Enabled: {enabled}")
            logger.info(f"  Primary: {primary}")
            logger.info(f"  Mode: {mode}")
            logger.info(f"  Position: {pos}")
            logger.info(f"  Refresh Rate: {str(refresh_rate) + 'hz' if refresh_rate else ''}")
            logger.info(f"  Primary: {primary}")
            logger.info(f"  Rotation: {rotation}")
            logger.info(f"  Scaling: {scaling}")

            
            
            
            xrandr_command = ['xrandr', '--output', display_name]

            if mode:
                xrandr_command.extend(['--mode', mode])
                
            if pos:
                xrandr_command.extend(['--pos', pos])
                
            if refresh_rate:
                xrandr_command.extend(['--rate', refresh_rate])
                
            if rotation:
                xrandr_command.extend(['--rotate', rotation])
                
            # Apply scaling override if provided
            if scaling:
                xrandr_command.extend(['--scale', scaling])
                
            if primary:
                xrandr_command.extend(['--primary'])

            # If display is primary it can't be disabled
            if not enabled and primary:
                logger.warning("You can't disable a primary display")
            
            if not enabled and not primary:
                xrandr_command.extend(['--off'])
            
            command = ' '.join(map(str, xrandr_command))
            logger.info(f"  Command: {command}")
            commands.append(command)
        return commands

def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description='Configure displays using xrandr')
    parser.add_argument('--config-file', default='config.toml', help='Path to the TOML config file')
    parser.add_argument('--dry-run', action='store_true', help='Print commands without applying changes')
    parser.add_argument('--generate', action='store_true', help='Generates a config based on the current xrandr setup and outputs it before exiting. Feel free to redirect the output where appropriate')
    parser.add_argument('--replace-plasmashell', action='store_true', default=False, help='Replaces plasmashell on completion to make sure changes apply globally')
    args = parser.parse_args()
    return args

def main() -> None:
    args = parse_args()
    strandr = STRandR()
    
    if args.generate:
        print(strandr.get_current_config()) # Only time we print because its designed to be redirected to a file
        return

    try:
        with open(args.config_file, 'r') as config_file:
            config = toml.load(config_file)
    except FileNotFoundError:
        logger.error(f"Error: Config file '{args.config_file}' not found (tip: try generating one from your current config with --generate)")
        return
    except toml.TomlDecodeError as e:
        logger.error(f"Error decoding config file: {e} (tip: try generating one from your current config with --generate)")
        return

    commands = strandr.configure_displays(config)
    for command in commands:
        if not command:
            logger.warn("Empty command")
        else:
            if not args.dry_run:
                logger.info(f"SHELL => '{command}'")
                subprocess.run(command, shell=True)

    if not args.dry_run and args.replace_plasmashell:
        logger.info("Restarting plasmashell")
        subprocess.run(['kquitapp5', 'plasmashell'], stdout = subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        subprocess.run(['kstart5', 'plasmashell'], stdout = subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        
    logger.info("Done!")


if __name__ == '__main__':
    main()